import	os
import	unittest
from	pygwarts.magical.philosophers_stone		import Transmutable
from	pygwarts.tests.irma						import IrmaTestCase
from	pygwarts.irma.contrib					import LibraryContrib
from	pygwarts.irma.shelve					import LibraryShelf
from	pygwarts.irma.shelve.casing				import is_num
from	pygwarts.irma.shelve.casing				import num_diff
from	pygwarts.irma.access.volume				import LibraryVolume
from	pygwarts.irma.access.bookmarks			import VolumeBookmark
from	pygwarts.irma.access.handlers			import AccessHandler
from	pygwarts.irma.access.inducers			import AccessInducer
from	pygwarts.irma.access.inducers.counters	import HandlerCounterInducer
from	pygwarts.irma.access.inducers.counters	import RegisterCounterInducer
from	pygwarts.irma.access.inducers.recap		import RegisterRecapInducer
from	pygwarts.irma.access.inducers.recap		import RegisterRecapAccumulatorInducer
from	pygwarts.irma.access.inducers.filters	import posnum
from	pygwarts.irma.access.inducers.filters	import negnum
from	pygwarts.irma.access.inducers.filters	import nonnegnum
from	pygwarts.irma.access.inducers.filters	import nonposnum
from	pygwarts.irma.access.inducers.filters	import zeronum
from	pygwarts.irma.access.inducers.filters	import plurnum
from	pygwarts.irma.access.inducers.case		import InducerCase








class AccessInducerCase(IrmaTestCase):

	"""
		Testing for fields and edge cases
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.ACCESS_INDUCER): os.remove(cls.ACCESS_INDUCER)

	@classmethod
	def setUpClass(cls): cls.make_loggy_file(cls, cls.ACCESS_INDUCER)
	def test_no_upper_init(self):

		class LonelyInducer(AccessInducer):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "no_upper_init"
				init_level	= 10

		with self.assertLogs("no_upper_init", 10) as case_loggy : self.test_case = LonelyInducer()
		self.assertIn(

			f"INFO:no_upper_init:AccessInducer {self.test_case} not registered to any handler",
			case_loggy.output
		)


	def test_handler_init(self):
		class LonelyHandler(AccessHandler):

			class LonelyInducer(AccessInducer):	pass
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "handler_init"
				init_level	= 10

		with self.assertLogs("handler_init", 10) as case_loggy : self.test_case = LonelyHandler()
		self.assertIn(

			f"INFO:handler_init:AccessHandler {self.test_case} not assigned to any bookmark",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:handler_init:AccessInducer {self.test_case.LonelyInducer} not assigned to any bookmark",
			case_loggy.output
		)


	def test_bookmark_init(self):

		class LonelyBookmark(VolumeBookmark):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "bookmark_init"
				init_level	= 10

			class LonelyHandler(AccessHandler):
				class LonelyInducer(AccessInducer):	pass

		with self.assertLogs("bookmark_init", 10) as case_loggy : self.test_case = LonelyBookmark()
		self.assertIn(

			f"INFO:bookmark_init:Bookmark {self.test_case} not assigned to library",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:bookmark_init:Assigned to {self.test_case} bookmark",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:bookmark_init:Assigned to {self.test_case} bookmark "
			f"as {self.test_case.LonelyHandler} inducer",
			case_loggy.output
		)


	def test_not_handler_init(self):

		class LonelyBookmark(VolumeBookmark):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "not_handler_init"
				init_level	= 10

			class NotHandler(Transmutable):
				class LonelyInducer(AccessInducer):	pass

		with self.assertLogs("not_handler_init", 10) as case_loggy : self.test_case = LonelyBookmark()
		self.assertIn(

			f"INFO:not_handler_init:Bookmark {self.test_case} not assigned to library",
			case_loggy.output
		)
		self.assertIn(

			"INFO:not_handler_init:AccessInducer "
			f"{self.test_case.NotHandler.LonelyInducer} not registered to any handler",
			case_loggy.output
		)


	def test_not_bookmark_init(self):

		class NotBookmark(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "not_bookmark_init"
				init_level	= 10

			class LonelyHandler(AccessHandler):
				class LonelyInducer(AccessInducer):	pass

		with self.assertLogs("not_bookmark_init", 10) as case_loggy : self.test_case = NotBookmark()
		self.assertIn(

			"INFO:not_bookmark_init:AccessInducer "
			f"{self.test_case.LonelyHandler.LonelyInducer} not assigned to any bookmark",
			case_loggy.output
		)




	def test_getters(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "getters"
				init_level	= 10

			class LonelyBookmark(VolumeBookmark):
				trigger = "trigger"

				class LonelyHandler(AccessHandler):
					class LonelyInducer(AccessInducer):	pass

		with self.assertLogs("getters", 10) as case_loggy : self.test_case = LonelyVolume()
		self.assertIn(

			f"INFO:getters:Volume {self.test_case} not assigned to any library",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:getters:Assigned to volume {self.test_case}", case_loggy.output)
		self.assertIn(

			f"DEBUG:getters:Assigned to {self.test_case.LonelyBookmark} bookmark",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:getters:Assigned to {self.test_case.LonelyBookmark} bookmark "
			f"as {self.test_case.LonelyBookmark.LonelyHandler} inducer",
			case_loggy.output
		)

		self.test_case.LonelyBookmark.LonelyHandler.access_handler_counter = 9001
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = {

			"counter":	420,
			"recap":	69,
		}
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.get_handler_counter(),
			9001
		)
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.get_register_counter(self.test_case),
			420
		)
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.get_register_recap(self.test_case),
			69
		)




	def test_handler_counter(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "handler_counter"
				init_level	= 10

			class LonelyBookmark(VolumeBookmark):
				trigger = "trigger"

				class LonelyHandler(AccessHandler):
					class LonelyInducer(HandlerCounterInducer):	pass

		with self.assertLogs("handler_counter", 10) as case_loggy : self.test_case = LonelyVolume()
		self.assertIn(

			f"INFO:handler_counter:Volume {self.test_case} not assigned to any library",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:handler_counter:Assigned to volume {self.test_case}", case_loggy.output)
		self.assertIn(

			f"DEBUG:handler_counter:Assigned to {self.test_case.LonelyBookmark} bookmark",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:handler_counter:Assigned to {self.test_case.LonelyBookmark} bookmark "
			f"as {self.test_case.LonelyBookmark.LonelyHandler} inducer",
			case_loggy.output
		)

		self.test_case.LonelyBookmark.LonelyHandler.access_handler_counter = 9001
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"9001"
		)

		self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = lambda E: E if 0 <E else None
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"9001"
		)

		self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = lambda E: E if 9000 <E else None
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"9001"
		)

		self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = lambda E: E if 0 == E else None
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))

		for	invalid in (

			1, 1., "filter", True, False, None, ...,
			[ lambda E : E if 0 <E else None ],
			( lambda E : E if 0 <E else None, ),
			{ lambda E : E if 0 <E else None },
			{ "filter": lambda E : E if 0 <E else None },
		):
			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = invalid
			with self.subTest(filter=invalid), self.assertLogs("handler_counter", 10) as case_loggy :
				self.assertEqual(

					self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
					"9001"
				)
			self.assertIn(f"DEBUG:handler_counter:Filter not applied", case_loggy.output)

		self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = None
		for	invalid in (

			"9001", None, ..., print, unittest,
			[ 9001 ],( 9001, ),{ 9001 },{ "access_handler_counter": 9001 },
		):
			self.test_case.LonelyBookmark.LonelyHandler.access_handler_counter = invalid
			with self.subTest(access_handler_counter=invalid):
				self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))








	def test_register_counter(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "register_counter"
				init_level	= 10

			class LonelyBookmark(VolumeBookmark):
				trigger = "trigger"

				class LonelyHandler(AccessHandler):
					class LonelyInducer(RegisterCounterInducer):	pass

		with self.assertLogs("register_counter", 10) as case_loggy : self.test_case = LonelyVolume()
		self.assertIn(

			f"INFO:register_counter:Volume {self.test_case} not assigned to any library",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:register_counter:Assigned to volume {self.test_case}", case_loggy.output)
		self.assertIn(

			f"DEBUG:register_counter:Assigned to {self.test_case.LonelyBookmark} bookmark",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:register_counter:Assigned to {self.test_case.LonelyBookmark} bookmark "
			f"as {self.test_case.LonelyBookmark.LonelyHandler} inducer",
			case_loggy.output
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "counter": 420 }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"420"
		)

		self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = lambda E : E if 0 <E else None
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"420"
		)

		self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = lambda E : E if 1 <E else None
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"420"
		)

		self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = lambda E : E if 0 == E else None
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))

		for	invalid in (

			1, 1., "filter", True, False, None, ...,
			[ lambda E : E if 0 <E else None ],
			( lambda E : E if 0 <E else None, ),
			{ lambda E : E if 0 <E else None },
			{ "filter": lambda E : E if 0 <E else None },
		):
			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = invalid
			with self.subTest(filter=invalid), self.assertLogs("register_counter", 10) as case_loggy :
				self.assertEqual(

					self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
					"420"
				)
			self.assertIn(f"DEBUG:register_counter:Filter not applied", case_loggy.output)

		self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = None
		for	invalid in (

			"420", None, ..., print, unittest,
			[ 420 ],( 420, ),{ 420 },{ "access_handler_counter": 420 },
		):
			self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "counter": invalid }
			with self.subTest(register_counter=invalid):
				self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))








	def test_register_recap(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "register_recap"
				init_level	= 10

			class LonelyBookmark(VolumeBookmark):
				trigger = "trigger"

				class LonelyHandler(AccessHandler):
					class LonelyInducer(RegisterRecapInducer):	pass

		with self.assertLogs("register_recap", 10) as case_loggy : self.test_case = LonelyVolume()
		self.assertIn(

			f"INFO:register_recap:Volume {self.test_case} not assigned to any library",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:register_recap:Assigned to volume {self.test_case}", case_loggy.output)
		self.assertIn(

			f"DEBUG:register_recap:Assigned to {self.test_case.LonelyBookmark} bookmark",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:register_recap:Assigned to {self.test_case.LonelyBookmark} bookmark "
			f"as {self.test_case.LonelyBookmark.LonelyHandler} inducer",
			case_loggy.output
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 69 }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"69"
		)

		self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = lambda E: E if 0 <E else None
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"69"
		)

		self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = lambda E: E if 1 <E else None
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"69"
		)

		self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = lambda E: E if 0 == E else None
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))

		for	invalid in (

			1, 1., "filter", True, False, None, ...,
			[ lambda E : E if 0 <E else None ],
			( lambda E : E if 0 <E else None, ),
			{ lambda E : E if 0 <E else None },
			{ "filter": lambda E : E if 0 <E else None },
		):
			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = invalid
			with self.subTest(filter=invalid), self.assertLogs("register_recap", 10) as case_loggy :
				self.assertEqual(

					self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
					"69"
				)
			self.assertIn(f"DEBUG:register_recap:Filter not applied", case_loggy.output)

		self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.filter = None
		for	variant in (

			69., "69", True, False, ...,
			[ 69 ],( 69, ),{ 69 },{ "recap": 69 },
		):
			with self.subTest(recap=variant):

				self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": variant }
				self.assertEqual(

					self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
					str(variant)
				)








	def test_register_recap_accumulator(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "recap_accumulator"
				init_level	= 10

			class LonelyBookmark(VolumeBookmark):
				trigger = "trigger"

				class LonelyHandler(AccessHandler):
					class LonelyInducer(RegisterRecapAccumulatorInducer):	pass

		with self.assertLogs("recap_accumulator", 10) as case_loggy:

			self.test_case = LonelyVolume()
			self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": [ 420,69,69 ] }
			self.assertEqual(

				self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
				"420 69 69"
			)

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.joint = "\n"
			self.assertEqual(

				self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
				"420\n69\n69"
			)

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer.unique = True
			self.assertEqual(

				self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
				"420\n69"
			)

		self.assertIn(

			f"INFO:recap_accumulator:Volume {self.test_case} not assigned to any library",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:recap_accumulator:Assigned to volume {self.test_case}", case_loggy.output)
		self.assertIn(

			f"DEBUG:recap_accumulator:Assigned to {self.test_case.LonelyBookmark} bookmark",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:recap_accumulator:Assigned to {self.test_case.LonelyBookmark} bookmark "
			f"as {self.test_case.LonelyBookmark.LonelyHandler} inducer",
			case_loggy.output
		)
		self.assertIn("DEBUG:recap_accumulator:Accumulated 9 symbols string", case_loggy.output)
		self.assertEqual(case_loggy.output.count("DEBUG:recap_accumulator:Accumulated 9 symbols string"), 2)
		self.assertIn("DEBUG:recap_accumulator:Accumulated 6 symbols string", case_loggy.output)

		for	invalid in (

			1, 1., "filter", True, False, None, ...,
			( 420,69,69, ),{ 420,69,69 },{ "filter": [ 420,69,69 ] }
		):
			self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": invalid }
			self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))








	def test_filter_posnum(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "filter_posnum"
				init_level	= 10

			class LonelyBookmark(VolumeBookmark):
				trigger = "trigger"

				class LonelyHandler(AccessHandler):
					class LonelyInducer(RegisterRecapInducer):	filter = posnum

		with self.assertLogs("filter_posnum", 10) as case_loggy : self.test_case = LonelyVolume()
		self.assertIn(

			f"INFO:filter_posnum:Volume {self.test_case} not assigned to any library",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:filter_posnum:Assigned to volume {self.test_case}", case_loggy.output)
		self.assertIn(

			f"DEBUG:filter_posnum:Assigned to {self.test_case.LonelyBookmark} bookmark",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:filter_posnum:Assigned to {self.test_case.LonelyBookmark} bookmark "
			f"as {self.test_case.LonelyBookmark.LonelyHandler} inducer",
			case_loggy.output
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 69 }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"69"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": .69 }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"0.69"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "69" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"69"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": ".69" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			".69"
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 0 }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 0. }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "0" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": ".0" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))

		for	invalid in ( "filter", True, False, None, ...,[ 42 ],( 42, ),{ 42 },{ "value": 42 }):
			with self.subTest(value=invalid):

				self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": invalid }
				self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))








	def test_filter_negnum(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "filter_negnum"
				init_level	= 10

			class LonelyBookmark(VolumeBookmark):
				trigger = "trigger"

				class LonelyHandler(AccessHandler):
					class LonelyInducer(RegisterRecapInducer):	filter = negnum

		with self.assertLogs("filter_negnum", 10) as case_loggy : self.test_case = LonelyVolume()
		self.assertIn(

			f"INFO:filter_negnum:Volume {self.test_case} not assigned to any library",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:filter_negnum:Assigned to volume {self.test_case}", case_loggy.output)
		self.assertIn(

			f"DEBUG:filter_negnum:Assigned to {self.test_case.LonelyBookmark} bookmark",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:filter_negnum:Assigned to {self.test_case.LonelyBookmark} bookmark "
			f"as {self.test_case.LonelyBookmark.LonelyHandler} inducer",
			case_loggy.output
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": -69 }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"-69"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": -.69 }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"-0.69"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "-69" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"-69"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "-.69" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"-.69"
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 0 }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 0. }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "0" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": ".0" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))

		for	invalid in ( "filter", True, False, None, ...,[ 42 ],( 42, ),{ 42 },{ "value": 42 }):
			with self.subTest(value=invalid):

				self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": invalid }
				self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))








	def test_filter_nonnegnum(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "filter_nonnegnum"
				init_level	= 10

			class LonelyBookmark(VolumeBookmark):
				trigger = "trigger"

				class LonelyHandler(AccessHandler):
					class LonelyInducer(RegisterRecapInducer):	filter = nonnegnum

		with self.assertLogs("filter_nonnegnum", 10) as case_loggy : self.test_case = LonelyVolume()
		self.assertIn(

			f"INFO:filter_nonnegnum:Volume {self.test_case} not assigned to any library",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:filter_nonnegnum:Assigned to volume {self.test_case}", case_loggy.output)
		self.assertIn(

			f"DEBUG:filter_nonnegnum:Assigned to {self.test_case.LonelyBookmark} bookmark",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:filter_nonnegnum:Assigned to {self.test_case.LonelyBookmark} bookmark "
			f"as {self.test_case.LonelyBookmark.LonelyHandler} inducer",
			case_loggy.output
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 0 }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"0"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 0. }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"0.0"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "0" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"0"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": ".0" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			".0"
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 69 }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"69"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 69. }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"69.0"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "69" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"69"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": ".69" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			".69"
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": -69 }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": -69. }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "-69" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "-.69" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))

		for	invalid in ( "filter", True, False, None, ...,[ 42 ],( 42, ),{ 42 },{ "value": 42 }):
			with self.subTest(value=invalid):

				self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": invalid }
				self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))








	def test_filter_nonposnum(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "filter_nonposnum"
				init_level	= 10

			class LonelyBookmark(VolumeBookmark):
				trigger = "trigger"

				class LonelyHandler(AccessHandler):
					class LonelyInducer(RegisterRecapInducer):	filter = nonposnum

		with self.assertLogs("filter_nonposnum", 10) as case_loggy : self.test_case = LonelyVolume()
		self.assertIn(

			f"INFO:filter_nonposnum:Volume {self.test_case} not assigned to any library",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:filter_nonposnum:Assigned to volume {self.test_case}", case_loggy.output)
		self.assertIn(

			f"DEBUG:filter_nonposnum:Assigned to {self.test_case.LonelyBookmark} bookmark",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:filter_nonposnum:Assigned to {self.test_case.LonelyBookmark} bookmark "
			f"as {self.test_case.LonelyBookmark.LonelyHandler} inducer",
			case_loggy.output
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 0 }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"0"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 0. }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"0.0"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "0" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"0"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": ".0" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			".0"
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": -69 }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"-69"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": -69. }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"-69.0"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "-69" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"-69"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "-.69" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"-.69"
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 69 }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 69. }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "69" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": ".69" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))

		for	invalid in ( "filter", True, False, None, ...,[ 42 ],( 42, ),{ 42 },{ "value": 42 }):
			with self.subTest(value=invalid):

				self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": invalid }
				self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))








	def test_filter_zeronum(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "filter_zeronum"
				init_level	= 10

			class LonelyBookmark(VolumeBookmark):
				trigger = "trigger"

				class LonelyHandler(AccessHandler):
					class LonelyInducer(RegisterRecapInducer):	filter = zeronum

		with self.assertLogs("filter_zeronum", 10) as case_loggy : self.test_case = LonelyVolume()
		self.assertIn(

			f"INFO:filter_zeronum:Volume {self.test_case} not assigned to any library",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:filter_zeronum:Assigned to volume {self.test_case}", case_loggy.output)
		self.assertIn(

			f"DEBUG:filter_zeronum:Assigned to {self.test_case.LonelyBookmark} bookmark",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:filter_zeronum:Assigned to {self.test_case.LonelyBookmark} bookmark "
			f"as {self.test_case.LonelyBookmark.LonelyHandler} inducer",
			case_loggy.output
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 0 }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"0"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 0. }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"0.0"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "0" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"0"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": ".0" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			".0"
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 69 }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 69. }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "69" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": ".69" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": -69 }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": -69. }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "-69" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "-.69" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))

		for	invalid in ( "filter", True, False, None, ...,[ 42 ],( 42, ),{ 42 },{ "value": 42 }):
			with self.subTest(value=invalid):

				self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": invalid }
				self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))








	def test_filter_plurnum(self):

		class LonelyVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "filter_plurnum"
				init_level	= 10

			class LonelyBookmark(VolumeBookmark):
				trigger = "trigger"

				class LonelyHandler(AccessHandler):
					class LonelyInducer(RegisterRecapInducer):	filter = plurnum

		with self.assertLogs("filter_plurnum", 10) as case_loggy : self.test_case = LonelyVolume()
		self.assertIn(

			f"INFO:filter_plurnum:Volume {self.test_case} not assigned to any library",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:filter_plurnum:Assigned to volume {self.test_case}", case_loggy.output)
		self.assertIn(

			f"DEBUG:filter_plurnum:Assigned to {self.test_case.LonelyBookmark} bookmark",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:filter_plurnum:Assigned to {self.test_case.LonelyBookmark} bookmark "
			f"as {self.test_case.LonelyBookmark.LonelyHandler} inducer",
			case_loggy.output
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 69 }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"69"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 69. }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"69.0"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "69" }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"69"
		)
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "69." }
		self.assertEqual(

			self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case),
			"69."
		)

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 1 }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 1. }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "1" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "1." }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))

		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 0 }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": 0. }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": "0" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))
		self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": ".0" }
		self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))

		for	invalid in ( "filter", True, False, None, ...,[ 42 ],( 42, ),{ 42 },{ "value": 42 }):
			with self.subTest(value=invalid):

				self.test_case[self.test_case.LonelyBookmark.LonelyHandler] = { "recap": invalid }
				self.assertIsNone(self.test_case.LonelyBookmark.LonelyHandler.LonelyInducer(self.test_case))








	def test_InducerCase(self):
		class FakeLibrary(Transmutable):

			class LonelyVolume(LibraryVolume):	pass
			class LonelyShelf(LibraryShelf):	pass
			class loggy(LibraryContrib):

				handler		= self.ACCESS_INDUCER
				init_name	= "InducerCase"
				init_level	= 10

			@InducerCase("LonelyShelf", prep=is_num, post=num_diff)
			class LonelyInducer(AccessInducer):
				def __call__(self, volume :LibraryVolume) -> str : return "69"

		self.test_case = FakeLibrary()
		self.assertIsNone(self.test_case.LonelyShelf[str(self.test_case.LonelyVolume)])
		self.assertEqual(self.test_case.LonelyInducer(self.test_case.LonelyVolume), "69 (+69)")
		self.assertEqual(self.test_case.LonelyInducer(self.test_case.LonelyVolume), "69")
		self.assertEqual(
			self.test_case.LonelyShelf[
				str(self.test_case.LonelyVolume)
			][	str(self.test_case.LonelyInducer)
			],	"69"
		)
		self.test_case.LonelyShelf[
				str(self.test_case.LonelyVolume)
			][	str(self.test_case.LonelyInducer)
			] =	"68"
		self.assertEqual(self.test_case.LonelyInducer(self.test_case.LonelyVolume), "69 (+1)")
		self.assertEqual(
			self.test_case.LonelyShelf[
				str(self.test_case.LonelyVolume)
			][	str(self.test_case.LonelyInducer)
			],	"69"
		)
		self.test_case.LonelyShelf[
				str(self.test_case.LonelyVolume)
			][	str(self.test_case.LonelyInducer)
			] =	"70"
		self.assertEqual(self.test_case.LonelyInducer(self.test_case.LonelyVolume), "69 (-1)")
		self.assertEqual(
			self.test_case.LonelyShelf[
				str(self.test_case.LonelyVolume)
			][	str(self.test_case.LonelyInducer)
			],	"69"
		)
		self.test_case.LonelyShelf[
				str(self.test_case.LonelyVolume)
			][	str(self.test_case.LonelyInducer)
			] =	"69"
		self.assertEqual(self.test_case.LonelyInducer(self.test_case.LonelyVolume), "69")
		self.assertEqual(
			self.test_case.LonelyShelf[
				str(self.test_case.LonelyVolume)
			][	str(self.test_case.LonelyInducer)
			],	"69"
		)








if	__name__ == "__main__" : unittest.main(verbosity=2)










from pygwarts.irma.access.bookmarks			import VolumeBookmark
from pygwarts.irma.access.handlers.counters	import AccessCounter
from pygwarts.irma.access.inducers.counters	import RegisterCounterInducer
from pygwarts.irma.access.inducers.filters	import posnum








class InfoCount(VolumeBookmark):

	""" INFO level messages counter. Induces only positive register counter. """

	trigger = "INFO : "

	class Counter(AccessCounter):
		class VolumeInducer(RegisterCounterInducer): filter = posnum








class WarningCount(VolumeBookmark):

	""" WARNING level messages counter. Induces only positive register counter.	"""

	trigger = "WARNING : "

	class Counter(AccessCounter):
		class VolumeInducer(RegisterCounterInducer): filter = posnum








class ErrorCount(VolumeBookmark):

	""" WARNING level messages counter. Induces only positive register counter.	"""

	trigger = "ERROR : "

	class Counter(AccessCounter):
		class VolumeInducer(RegisterCounterInducer): filter = posnum








class CriticalCount(VolumeBookmark):

	""" WARNING level messages counter. Induces only positive register counter.	"""

	trigger = "CRITICAL : "

	class Counter(AccessCounter):
		class VolumeInducer(RegisterCounterInducer): filter = posnum










import	unittest
from	math					import inf
from	pygwarts.magical.spells	import flagrate
from	pygwarts.tests.magical	import MagicalTestCase








class SpellsCases(MagicalTestCase):
	def test_flagrate_singular(self):

		self.assertEqual(flagrate(1),"")
		self.assertEqual(flagrate(-1),"")
		self.assertEqual(flagrate("1"),"")
		self.assertEqual(flagrate("-1"),"")
		self.assertEqual(flagrate(""),"")
		self.assertEqual(flagrate("True"),"")
		self.assertEqual(flagrate("False"),"")
		self.assertEqual(flagrate("None"),"")
		self.assertEqual(flagrate(True),"")
		self.assertEqual(flagrate(None),"")
		self.assertEqual(flagrate([ 42 ]),"")
		self.assertEqual(flagrate(( 42, )),"")
		self.assertEqual(flagrate({ 42, }),"")
		self.assertEqual(flagrate({ 42: 42, }),"")
		self.assertEqual(flagrate(print),"")
		self.assertEqual(flagrate(...),"")
		self.assertEqual(flagrate(self),"")
		self.assertEqual(flagrate(ValueError),"")

	def test_flagrate_plural(self):

		self.assertEqual(flagrate(False),"s")
		self.assertEqual(flagrate(0),"s")
		self.assertEqual(flagrate(0.),"s")
		self.assertEqual(flagrate(.0),"s")
		self.assertEqual(flagrate(-0),"s")
		self.assertEqual(flagrate(-0.),"s")
		self.assertEqual(flagrate(-.0),"s")
		self.assertEqual(flagrate("0"),"s")
		self.assertEqual(flagrate("0."),"s")
		self.assertEqual(flagrate(".0"),"s")
		self.assertEqual(flagrate("-0"),"s")
		self.assertEqual(flagrate("-0."),"s")
		self.assertEqual(flagrate("-.0"),"s")
		self.assertEqual(flagrate(1.),"s")
		self.assertEqual(flagrate(-1.),"s")
		self.assertEqual(flagrate("1."),"s")
		self.assertEqual(flagrate("-1."),"s")
		self.assertEqual(flagrate(2),"s")
		self.assertEqual(flagrate(2.),"s")
		self.assertEqual(flagrate(10),"s")
		self.assertEqual(flagrate(10.1),"s")
		self.assertEqual(flagrate(100500),"s")
		self.assertEqual(flagrate(-2),"s")
		self.assertEqual(flagrate(-2.),"s")
		self.assertEqual(flagrate(-10),"s")
		self.assertEqual(flagrate(-10.1),"s")
		self.assertEqual(flagrate(-100500),"s")
		self.assertEqual(flagrate("2"),"s")
		self.assertEqual(flagrate("2."),"s")
		self.assertEqual(flagrate("10"),"s")
		self.assertEqual(flagrate("10.1"),"s")
		self.assertEqual(flagrate("100500"),"s")
		self.assertEqual(flagrate("-2"),"s")
		self.assertEqual(flagrate("-2."),"s")
		self.assertEqual(flagrate("-10"),"s")
		self.assertEqual(flagrate("-10.1"),"s")
		self.assertEqual(flagrate("-100500"),"s")
		self.assertEqual(flagrate(1E1),"s")
		self.assertEqual(flagrate(2E200),"s")
		self.assertEqual(flagrate(-1E1),"s")
		self.assertEqual(flagrate(-2E200),"s")
		self.assertEqual(flagrate("1E1"),"s")
		self.assertEqual(flagrate("2E200"),"s")
		self.assertEqual(flagrate("-1E1"),"s")
		self.assertEqual(flagrate("-2E200"),"s")
		self.assertEqual(flagrate(inf),"s")
		self.assertEqual(flagrate(-inf),"s")
		self.assertEqual(flagrate("inf"),"s")
		self.assertEqual(flagrate("-inf"),"s")








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	typing								import Hashable
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.tests.irma					import IrmaTestCase
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.irma.shelve				import LibraryShelf
from	pygwarts.irma.shelve.casing			import shelf_case
from	pygwarts.irma.shelve.casing			import is_num
from	pygwarts.irma.shelve.casing			import num_diff
from	pygwarts.irma.shelve.casing			import seq_diff
from	pygwarts.irma.shelve.casing			import ShelfCase
from	pygwarts.irma.shelve.casing			import NumDiffCase
from	pygwarts.irma.shelve.casing			import SeqDiffCase








class ShelfCasingCase(IrmaTestCase):

	"""
		Testing for fields and edge cases
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.CASING_HANDLER): os.remove(cls.CASING_HANDLER)

	@classmethod
	def setUpClass(cls):

		super().setUpClass()
		cls.make_loggy_file(cls, cls.CASING_HANDLER)
		cls.prep1 = lambda _,A	: A
		cls.prep2 = lambda _,A	: f"{A} bananas"
		cls.post1 = lambda _,A,B: str(A)
		cls.post2 = lambda _,A,B: f"It's {A} and so on"
		cls.current_shelf = LibraryShelf()




	def test_shelf_case_targets(self):
		for target in (

			42, 69., "ten", True, False, ..., int, Transmutable,[ 1 ],( 1, ),{ 1 },{ "value": 1 }
		):
			with self.subTest(target=target):

				self.assertEqual(
					shelf_case(

						target,
						key="test",
						shelf=self.current_shelf,
						prep=self.prep1,
						post=self.post1,
					),	str(target)
				)
				self.assertEqual(self.current_shelf["test"], target)

		self.assertIsNotNone(self.current_shelf["test"])
		self.assertIsNone(
			shelf_case(

				None,
				key="test",
				shelf=self.current_shelf,
				prep=self.prep1,
				post=self.post1,
			),	str(target)
		)
		self.assertIsNotNone(self.current_shelf["test"])




	def test_shelf_case_keys(self):

		for item in ( 42, 69., "ten", True, False, None, ..., int, Transmutable,( 1, )):
			with self.subTest(key=item):

				self.assertEqual(
					shelf_case(

						"target",
						key=item,
						shelf=self.current_shelf,
						prep=self.prep1,
						post=self.post1,
					),	"target"
				)
				self.assertEqual(self.current_shelf[item], "target")




	def test_invalid_key_shelf_case(self):

		for invalid in ([ 1 ],{ 1 },{ "value": 1 }):
			with self.subTest(key=invalid):

				self.assertNotIsInstance(invalid, Hashable)
				self.assertIsNone(
					shelf_case(

						"target",
						key=invalid,
						shelf=self.current_shelf,
						prep=self.prep1,
						post=self.post1,
					)
				)




	def test_shelf_case_invalid_shelf(self):
		for invalid in (

			42, 69., "ten", True, False, None, ..., int, Transmutable,[ 1 ],( 1, ),{ 1 }
		):
			with self.subTest(shelf=invalid):

				self.assertNotIsInstance(invalid, LibraryShelf)
				self.assertIsNone(
					shelf_case(

						"target",
						key="test",
						shelf=invalid,
						prep=self.prep1,
						post=self.post1,
					)
				)




	def test_shelf_case_invalid_prep(self):
		for invalid in (

			42, 69., "ten", True, False, None, ..., super, Transmutable,[ 1 ],( 1, ),{ 1 },{ "value": 1 }
		):
			with self.subTest(prep=invalid):
				self.assertIsNone(
					shelf_case(

						42,
						key="test",
						shelf=self.current_shelf,
						prep=invalid,
						post=self.post1,
					)
				)




	def test_shelf_case_invalid_post(self):
		for invalid in (

			42, 69., "ten", True, False, None, ..., super, Transmutable,[ 1 ],( 1, ),{ 1 },{ "value": 1 }
		):
			with self.subTest(post=invalid):
				self.assertIsNone(
					shelf_case(

						42,
						key="test",
						shelf=self.current_shelf,
						prep=self.prep1,
						post=invalid,
					)
				)








	def test_is_num_int_prep(self):

		for i in range(-100,101,1):
			with self.subTest(candidate=i): self.assertEqual(is_num(i),i)


	def test_is_num_str_prep(self):

		for i in range(-100,101,1):
			j = str(i)
			with self.subTest(candidate=j): self.assertEqual(is_num(j),j)


	def test_is_num_float_prep(self):

		for i in range(101):
			j = float(f".{i}")
			k = -j

			with self.subTest(candidates=( j,k )):

				self.assertEqual(is_num(j),j)
				self.assertEqual(is_num(k),k)








	def test_valid_num_diff_post(self):

		for i in range(101):
			for j in range(i +1,102):

				r00 = str(j)
				r01 = float(j)
				r02 = float(i)
				r03 = float(j -i)
				r04 = float(i -j)
				r05 = f"{j} (+{j -i})"
				r06 = f"{j} (+{r03})"
				r07 = f"{r01} (+{r03})"
				r08 = f"{i} ({i -j})"
				r09 = f"{i} ({r04})"
				r10 = f"{r02} ({r04})"

				self.assertEqual(num_diff(j,i),					r05)
				self.assertEqual(num_diff(j,str(i)),			r05)
				self.assertEqual(num_diff(j,float(i)),			r06)
				self.assertEqual(num_diff(str(j),i),			r05)
				self.assertEqual(num_diff(str(j),str(i)),		r05)
				self.assertEqual(num_diff(str(j),float(i)),		r06)
				self.assertEqual(num_diff(float(j),i),			r07)
				self.assertEqual(num_diff(float(j),str(i)),		r07)
				self.assertEqual(num_diff(float(j),float(i)),	r07)

				self.assertEqual(num_diff(i,j),					r08)
				self.assertEqual(num_diff(i,str(j)),			r08)
				self.assertEqual(num_diff(i,float(j)),			r09)
				self.assertEqual(num_diff(str(i),j),			r08)
				self.assertEqual(num_diff(str(i),str(j)),		r08)
				self.assertEqual(num_diff(str(i),float(j)),		r09)
				self.assertEqual(num_diff(float(i),j),			r10)
				self.assertEqual(num_diff(float(i),str(j)),		r10)
				self.assertEqual(num_diff(float(i),float(j)),	r10)

				self.assertEqual(num_diff(j,j),					r00)
				self.assertEqual(num_diff(j,str(j)),			r00)
				self.assertEqual(num_diff(j,float(j)),			r00)
				self.assertEqual(num_diff(str(j),j),			r00)
				self.assertEqual(num_diff(str(j),str(j)),		r00)
				self.assertEqual(num_diff(str(j),float(j)),		r00)
				self.assertEqual(num_diff(float(j),j),			str(r01))
				self.assertEqual(num_diff(float(j),str(j)),		str(r01))
				self.assertEqual(num_diff(float(j),float(j)),	str(r01))


	def test_invalid_num_diff_post(self):

		for invalid in ( ..., [ 1 ],( 1, ),{ 1 },{ "num1": 1 }):
			with self.subTest(num=invalid):

				self.assertRaises(TypeError, num_diff, 42, invalid)
				self.assertRaises(TypeError, num_diff, invalid, 42)
				if	invalid != { 1 } : self.assertRaises(TypeError, num_diff, invalid, invalid)

		for invalid in ( print, Transmutable ):
			with self.subTest(num=invalid):

				self.assertRaises(SyntaxError, num_diff, 42, invalid)
				self.assertRaises(SyntaxError, num_diff, invalid, 42)
				self.assertRaises(SyntaxError, num_diff, invalid, invalid)

		for invalid in ( "one", ):
			with self.subTest(num=invalid):

				self.assertRaises(NameError, num_diff, 42, invalid)
				self.assertRaises(NameError, num_diff, invalid, 42)
				self.assertRaises(NameError, num_diff, invalid, invalid)








	def test_valid_same_seq_diff_post(self):
		for seq1 in (

			( 1,2,3,4,5,6 ),[ 1,2,3,4,5,6 ],{ 1,2,3,4,5,6 },
			{ 1: True, 2: True, 3: True, 4: True, 5: True, 6: True },
		):
			for seq2 in (

				( 1,2,3,4 ),[ 1,2,3,4 ],{ 1,2,3,4 },
				{ 1: True, 2: True, 3: True, 4: True },
			):
				with self.subTest(seq1=seq1, seq2=seq2):
					self.assertCountEqual(seq_diff(seq1, seq2),[ "1","2","3","4","5 (+)","6 (+)" ])

				with self.subTest(seq1=seq1, seq2=seq1):
					self.assertCountEqual(seq_diff(seq1, seq1),[ "1","2","3","4","5","6" ])

				with self.subTest(seq1=seq2, seq2=seq1):
					self.assertCountEqual(seq_diff(seq2, seq1),[ "1","2","3","4" ])

				with self.subTest(seq1=seq2, seq2=seq2):
					self.assertCountEqual(seq_diff(seq2, seq1),[ "1","2","3","4" ])




	def test_valid_different_seq_diff_post(self):
		for seq1 in (

			( 1,2.,"3",4,5.,"6" ),[ 1,2.,"3",4,5.,"6" ],{ 1,2.,"3",4,5.,"6" },
			{ 1: True, 2.: True, "3": True, 4: True, 5.: True, "6": True },
		):
			for seq2 in (

				( "1",2,3,4 ),[ "1",2,3,4 ],{ "1",2,3,4 },
				{ "1": True, 2: True, 3: True, 4: True },
			):
				with self.subTest(seq1=seq1, seq2=seq2):
					self.assertCountEqual(
						seq_diff(seq1, seq2),[ "1 (+)","2.0","3 (+)","4","5.0 (+)","6 (+)" ]
					)

				with self.subTest(seq1=seq1, seq2=seq1):
					self.assertCountEqual(seq_diff(seq1, seq1),[ "1","2.0","3","4","5.0","6" ])

				with self.subTest(seq1=seq2, seq2=seq1):
					self.assertCountEqual(seq_diff(seq2, seq1),[ "1 (+)","2","3 (+)","4" ])

				with self.subTest(seq1=seq2, seq2=seq2):
					self.assertCountEqual(seq_diff(seq2, seq2),[ "1","2","3","4" ])




	def test_invalid_seq_diff_post(self):

		valid_seq = ( 1,2,3,4,5 )
		for invalid in ( 42, 69., True, ..., int, Transmutable ):

			self.assertRaises(TypeError, seq_diff, invalid, valid_seq)
			self.assertRaises(TypeError, seq_diff, valid_seq, invalid)
			self.assertRaises(TypeError, seq_diff, invalid, invalid)








	def test_ShelfCase_targets(self):

		@ShelfCase("CurrentShelf")
		class Shelfinator(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.CASING_HANDLER
				init_name	= "ShelfCase_targets"
				init_level	= 10

			class CurrentShelf(LibraryShelf):	pass
			def __call__(self, KEY): return KEY

		self.test_case = Shelfinator()

		with self.assertLogs("ShelfCase_targets", 10) as case_loggy:
			for target in (

				42, 69., "ten", True, False, ..., int, Transmutable,[ 1 ],( 1, ),{ 1 },{ "value": 1 }
			):
				with self.subTest(target=target):

					self.assertEqual(self.test_case(target), target)
					self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], target)

			self.assertIn(
				f"DEBUG:ShelfCase_targets:Obtained shelf casing value \"{target}\"", case_loggy.output
			)




	def test_invalid_links_ShelfCase(self):
		for i,invalid in enumerate((

			42, 69., "ten", True, False, None, ..., int, Transmutable,[ 1 ],( 1, ),{ 1 }
		)):
			@ShelfCase(invalid)
			class Shelfinator(Transmutable):
				class loggy(LibraryContrib):

					handler		= self.CASING_HANDLER
					init_name	= f"invalid_links_ShelfCase-{i}"
					init_level	= 10

				class CurrentShelf(LibraryShelf):	pass
				def __call__(self): return "cased"

			self.test_case = Shelfinator()

			with self.assertLogs(f"invalid_links_ShelfCase-{i}", 10) as case_loggy:
				self.assertEqual(self.test_case(), "cased")

			self.assertIn(

				f"DEBUG:invalid_links_ShelfCase-{i}:Failed to obtain shelf casing value",
				case_loggy.output
			)
			self.test_case.loggy.close()




	def test_ShelfCase_prep_targets(self):

		@ShelfCase("CurrentShelf", prep=self.prep2)
		class Shelfinator(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.CASING_HANDLER
				init_name	= "ShelfCase_prep_targets"
				init_level	= 10

			class CurrentShelf(LibraryShelf):	pass
			def __call__(self, KEY): return KEY

		self.test_case = Shelfinator()

		with self.assertLogs("ShelfCase_prep_targets", 10) as case_loggy:
			for target in (

				42, 69., "ten", True, False, ..., int, Transmutable,[ 1 ],( 1, ),{ 1 },{ "value": 1 }
			):
				with self.subTest(target=target):

					self.assertEqual(self.test_case(target), f"{target} bananas")
					self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], f"{target} bananas")

			self.assertIn(

				f"DEBUG:ShelfCase_prep_targets:Obtained shelf casing value \"{target} bananas\"",
				case_loggy.output
			)




	def test_ShelfCase_post_targets(self):

		@ShelfCase("CurrentShelf", post=self.post2)
		class Shelfinator(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.CASING_HANDLER
				init_name	= "ShelfCase_post_targets"
				init_level	= 10

			class CurrentShelf(LibraryShelf):	pass
			def __call__(self, KEY): return KEY

		self.test_case = Shelfinator()

		with self.assertLogs("ShelfCase_post_targets", 10) as case_loggy:
			for target in (

				42, 69., "ten", True, False, ..., int, Transmutable,[ 1 ],( 1, ),{ 1 },{ "value": 1 }
			):
				with self.subTest(target=target):

					self.assertEqual(self.test_case(target), f"It's {target} and so on")
					self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], target)

			self.assertIn(

				f"DEBUG:ShelfCase_post_targets:Obtained shelf casing value \"It's {target} and so on\"",
				case_loggy.output
			)




	def test_ShelfCase_prep_post_targets(self):

		@ShelfCase("CurrentShelf", prep=self.prep2, post=self.post2)
		class Shelfinator(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.CASING_HANDLER
				init_name	= "ShelfCase_prep_post_targets"
				init_level	= 10

			class CurrentShelf(LibraryShelf):	pass
			def __call__(self, KEY): return KEY

		self.test_case = Shelfinator()

		with self.assertLogs("ShelfCase_prep_post_targets", 10) as case_loggy:
			for target in (

				42, 69., "ten", True, False, ..., int, Transmutable,[ 1 ],( 1, ),{ 1 },{ "value": 1 }
			):
				with self.subTest(target=target):

					self.assertEqual(self.test_case(target), f"It's {target} bananas and so on")
					self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], f"{target} bananas")

			self.assertIn(

				"DEBUG:ShelfCase_prep_post_targets:"
				f"Obtained shelf casing value \"It's {target} bananas and so on\"",
				case_loggy.output
			)




	def test_ShelfCase_valid_prep_invalid_post_targets(self):
		for i,invalid in enumerate((

			42, 69., "ten", True, False, None, ..., super, Transmutable,[ 1 ],( 1, ),{ 1 },{ "value": 1 }
		)):
			with self.subTest(post=invalid):

				@ShelfCase("CurrentShelf", prep=self.prep2, post=invalid)
				class Shelfinator(Transmutable):
					class loggy(LibraryContrib):

						handler		= self.CASING_HANDLER
						init_name	= f"ShelfCase_valid_prep_invalid_post-{i}"
						init_level	= 10

					class CurrentShelf(LibraryShelf):	pass
					def __call__(self): return "forty two"

				self.test_case = Shelfinator()

				with self.assertLogs(f"ShelfCase_valid_prep_invalid_post-{i}", 10) as case_loggy:

					self.assertEqual(self.test_case(), "forty two")
					self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], "forty two bananas")

				self.assertIn(

					f"DEBUG:ShelfCase_valid_prep_invalid_post-{i}:Failed to obtain shelf casing value",
					case_loggy.output
				)
				self.test_case.loggy.close()




	def test_ShelfCase_invalid_prep_valid_post_targets(self):
		for i,invalid in enumerate((

			42, 69., "ten", True, False, None, ..., super, Transmutable,[ 1 ],( 1, ),{ 1 },{ "value": 1 }
		)):
			with self.subTest(prep=invalid):

				@ShelfCase("CurrentShelf", prep=invalid, post=self.post2)
				class Shelfinator(Transmutable):
					class loggy(LibraryContrib):

						handler		= self.CASING_HANDLER
						init_name	= f"ShelfCase_invalid_prep_valid_post-{i}"
						init_level	= 10

					class CurrentShelf(LibraryShelf):	pass
					def __call__(self): return "forty two"

				self.test_case = Shelfinator()

				with self.assertLogs(f"ShelfCase_invalid_prep_valid_post-{i}", 10) as case_loggy:

					self.assertEqual(self.test_case(), "forty two")
					self.assertIsNone(self.test_case.CurrentShelf[str(self.test_case)])

				self.assertIn(

					f"DEBUG:ShelfCase_invalid_prep_valid_post-{i}:Failed to obtain shelf casing value",
					case_loggy.output
				)
				self.test_case.loggy.close()




	def test_ShelfCase_invalid_prep_invalid_post_targets(self):
		for i,invalid in enumerate((

			42, 69., "ten", True, False, None, ..., super, Transmutable,[ 1 ],( 1, ),{ 1 },{ "value": 1 }
		)):
			with self.subTest(prep=invalid, post=invalid):

				@ShelfCase("CurrentShelf", prep=invalid, post=invalid)
				class Shelfinator(Transmutable):
					class loggy(LibraryContrib):

						handler		= self.CASING_HANDLER
						init_name	= f"ShelfCase_invalid_prep_invalid_post-{i}"
						init_level	= 10

					class CurrentShelf(LibraryShelf):	pass
					def __call__(self): return "forty two"

				self.test_case = Shelfinator()

				with self.assertLogs(f"ShelfCase_invalid_prep_invalid_post-{i}", 10) as case_loggy:

					self.assertEqual(self.test_case(), "forty two")
					self.assertIsNone(self.test_case.CurrentShelf[str(self.test_case)])

				self.assertIn(

					f"DEBUG:ShelfCase_invalid_prep_invalid_post-{i}:Failed to obtain shelf casing value",
					case_loggy.output
				)
				self.test_case.loggy.close()








	def test_NumDiffCase_int(self):

		@NumDiffCase("CurrentShelf")
		class Numericator(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.CASING_HANDLER
				init_name	= "NumDiffCase_int"
				init_level	= 10

			class CurrentShelf(LibraryShelf):	pass
			def __call__(self): return 42

		self.test_case = Numericator()
		self.test_case.CurrentShelf[str(self.test_case)] = 0
		self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 0)

		with self.assertLogs("NumDiffCase_int", 10) as case_loggy:

			self.assertEqual(self.test_case(), "42 (+42)")
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 42)
			self.assertEqual(self.test_case(), "42")
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 42)
			self.test_case.CurrentShelf[str(self.test_case)] = 84
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 84)
			self.assertEqual(self.test_case(), "42 (-42)")
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 42)

		self.assertIn(
			"DEBUG:NumDiffCase_int:Obtained number difference string \"42 (+42)\"", case_loggy.output
		)
		self.assertIn(
			"DEBUG:NumDiffCase_int:Obtained number difference string \"42\"", case_loggy.output
		)
		self.assertIn(
			"DEBUG:NumDiffCase_int:Obtained number difference string \"42 (-42)\"", case_loggy.output
		)




	def test_NumDiffCase_float(self):

		@NumDiffCase("CurrentShelf")
		class Numericator(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.CASING_HANDLER
				init_name	= "NumDiffCase_float"
				init_level	= 10

			class CurrentShelf(LibraryShelf):	pass
			def __call__(self): return 69.

		self.test_case = Numericator()
		self.test_case.CurrentShelf[str(self.test_case)] = 0
		self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 0)

		with self.assertLogs("NumDiffCase_float", 10) as case_loggy:

			self.assertEqual(self.test_case(), "69.0 (+69.0)")
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 69.0)
			self.assertEqual(self.test_case(), "69.0")
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 69.0)
			self.test_case.CurrentShelf[str(self.test_case)] = 138
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 138)
			self.assertEqual(self.test_case(), "69.0 (-69.0)")
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 69.0)

		self.assertIn(
			"DEBUG:NumDiffCase_float:Obtained number difference string \"69.0 (+69.0)\"", case_loggy.output
		)
		self.assertIn(
			"DEBUG:NumDiffCase_float:Obtained number difference string \"69.0\"", case_loggy.output
		)
		self.assertIn(
			"DEBUG:NumDiffCase_float:Obtained number difference string \"69.0 (-69.0)\"", case_loggy.output
		)




	def test_NumDiffCase_int_str(self):

		@NumDiffCase("CurrentShelf")
		class Numericator(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.CASING_HANDLER
				init_name	= "NumDiffCase_int_str"
				init_level	= 10

			class CurrentShelf(LibraryShelf):	pass
			def __call__(self): return "42"

		self.test_case = Numericator()
		self.test_case.CurrentShelf[str(self.test_case)] = 0
		self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 0)

		with self.assertLogs("NumDiffCase_int_str", 10) as case_loggy:

			self.assertEqual(self.test_case(), "42 (+42)")
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], "42")
			self.assertEqual(self.test_case(), "42")
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], "42")
			self.test_case.CurrentShelf[str(self.test_case)] = 84
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 84)
			self.assertEqual(self.test_case(), "42 (-42)")
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], "42")

		self.assertIn(
			"DEBUG:NumDiffCase_int_str:Obtained number difference string \"42 (+42)\"", case_loggy.output
		)
		self.assertIn(
			"DEBUG:NumDiffCase_int_str:Obtained number difference string \"42\"", case_loggy.output
		)
		self.assertIn(
			"DEBUG:NumDiffCase_int_str:Obtained number difference string \"42 (-42)\"", case_loggy.output
		)




	def test_NumDiffCase_float_str(self):

		@NumDiffCase("CurrentShelf")
		class Numericator(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.CASING_HANDLER
				init_name	= "NumDiffCase_float_str"
				init_level	= 10

			class CurrentShelf(LibraryShelf):	pass
			def __call__(self): return "69."

		self.test_case = Numericator()
		self.test_case.CurrentShelf[str(self.test_case)] = 0
		self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 0)

		with self.assertLogs("NumDiffCase_float_str", 10) as case_loggy:

			self.assertEqual(self.test_case(), "69. (+69.0)")
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], "69.")
			self.assertEqual(self.test_case(), "69.")
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], "69.")
			self.test_case.CurrentShelf[str(self.test_case)] = 138
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 138)
			self.assertEqual(self.test_case(), "69. (-69.0)")
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], "69.")

		self.assertIn(

			"DEBUG:NumDiffCase_float_str:Obtained number difference string \"69. (+69.0)\"",
			case_loggy.output
		)
		self.assertIn(

			"DEBUG:NumDiffCase_float_str:Obtained number difference string \"69.\"",
			case_loggy.output
		)
		self.assertIn(

			"DEBUG:NumDiffCase_float_str:Obtained number difference string \"69. (-69.0)\"",
			case_loggy.output
		)




	def test_not_num_NumDiffCase(self):

		@NumDiffCase("CurrentShelf")
		class Numericator(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.CASING_HANDLER
				init_name	= "not_num_NumDiffCase"
				init_level	= 10

			class CurrentShelf(LibraryShelf):	pass
			def __call__(self, KEY): return KEY

		self.test_case = Numericator()
		self.test_case.CurrentShelf[str(self.test_case)] = 0
		self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 0)

		with self.assertLogs("not_num_NumDiffCase", 10) as case_loggy:
			for invalid in (

				"ten", True, False, None, ..., print, Transmutable,[ 1 ],( 1, ),{ 1 },{ "target": 1 }
			):
				with self.subTest(target=invalid): self.assertEqual(self.test_case(invalid), invalid)

		self.assertIn("DEBUG:not_num_NumDiffCase:Number difference not obtained", case_loggy.output)
		self.assertEqual(
			# must be exactly 10, cause "None" doesn't trigger loggy
			case_loggy.output.count("DEBUG:not_num_NumDiffCase:Number difference not obtained"), 10
		)




	def test_invalid_link_NumDiffCase(self):
		for i,invalid in enumerate((

			42, 69., "ten", True, False, None, ..., print, Transmutable,[ 1 ],( 1, ),{ 1 },{ "target": 1 }
		)):
			with self.subTest(link=invalid):

				@NumDiffCase(invalid)
				class Numericator(Transmutable):
					class loggy(LibraryContrib):

						handler		= self.CASING_HANDLER
						init_name	= f"invalid_link_NumDiffCase-{i}"
						init_level	= 10

					class CurrentShelf(LibraryShelf):	pass
					def __call__(self): return 42

				self.test_case = Numericator()
				self.test_case.CurrentShelf[str(self.test_case)] = 0

				with self.assertLogs(f"invalid_link_NumDiffCase-{i}", 10) as case_loggy:
					self.assertEqual(self.test_case(), 42)

				self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], 0)
				self.assertIn(

					f"DEBUG:invalid_link_NumDiffCase-{i}:Number difference not obtained",
					case_loggy.output
				)
				self.test_case.loggy.close()








	def test_valid_SeqDiffCase(self):

		@SeqDiffCase("CurrentShelf")
		class Sequenator(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.CASING_HANDLER
				init_name	= "valid_SeqDiffCase"
				init_level	= 10

			class CurrentShelf(LibraryShelf):	pass
			def __call__(self, KEY): return KEY

		self.test_case = Sequenator()
		seq1 = [ 1, 3, 2, 4 ]
		seq2 = ( 1, 1, 3, 2, 2, 3, 4, 5, 6, 6 )
		res1 = [ "1 (+)", "3 (+)", "2 (+)", "4 (+)" ]
		res2 = [ "1", "1", "3", "2", "2", "3", "4", "5 (+)", "6 (+)", "6 (+)" ]

		with self.assertLogs("valid_SeqDiffCase", 10) as case_loggy:

			self.assertEqual(self.test_case(seq1), res1)
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], seq1)
			self.assertEqual(self.test_case(seq2), res2)
			self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)], seq2)

		self.assertIn(f"DEBUG:valid_SeqDiffCase:Obtained sequence difference \"{res1}\"", case_loggy.output)
		self.assertIn(f"DEBUG:valid_SeqDiffCase:Obtained sequence difference \"{res2}\"", case_loggy.output)




	def test_invalid_SeqDiffCase(self):

		@SeqDiffCase("CurrentShelf")
		class Sequenator(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.CASING_HANDLER
				init_name	= "invalid_SeqDiffCase"
				init_level	= 10

			class CurrentShelf(LibraryShelf):	pass
			def __call__(self, KEY): return KEY

		self.test_case = Sequenator()

		with self.assertLogs("invalid_SeqDiffCase", 10) as case_loggy:
			for invalid in ( 42, 69., True, False, None, ..., print, Transmutable ):
				with self.subTest(seq1=invalid):

					self.assertEqual(self.test_case(invalid), invalid)
					self.assertIsNone(self.test_case.CurrentShelf[str(self.test_case)])

		self.assertIn(f"DEBUG:invalid_SeqDiffCase:Sequence difference not obtained", case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"DEBUG:invalid_SeqDiffCase:Sequence difference not obtained"), 7
		)




	def test_invalid_link_SeqDiffCase(self):
		for i,invalid in enumerate((

			42, 69., "ten", True, False, None, ..., print, Transmutable,[ 1 ],( 1, ),{ 1 },{ "target": 1 }
		)):
			with self.subTest(link=invalid):

				@SeqDiffCase(invalid)
				class Sequenator(Transmutable):
					class loggy(LibraryContrib):

						handler		= self.CASING_HANDLER
						init_name	= f"invalid_link_SeqDiffCase-{i}"
						init_level	= 10

					class CurrentShelf(LibraryShelf):	pass
					def __call__(self): return [ 42 ]

				self.test_case = Sequenator()
				self.test_case.CurrentShelf[str(self.test_case)] = [ 0 ]

				with self.assertLogs(f"invalid_link_SeqDiffCase-{i}", 10) as case_loggy:
					self.assertEqual(self.test_case(),[ 42 ])

				self.assertEqual(self.test_case.CurrentShelf[str(self.test_case)],[ 0 ])
				self.assertIn(

					f"DEBUG:invalid_link_SeqDiffCase-{i}:Sequence difference not obtained",
					case_loggy.output
				)
				self.test_case.loggy.close()








if	__name__ == "__main__" : unittest.main(verbosity=2)










from __future__					import annotations
from typing						import Optional
from pygwarts.magical.spells	import patronus
from pygwarts.magical.spells	import wingardium_leviosa
from pygwarts.magical.spells	import evanesce
from pygwarts.irma.contrib		import LibraryContrib








class Transmutable:

	"""
                                                                           ************+=+++**=       
                                                           :****************#####**+==---+++**=       
                                                       -***###*****+********#**+=-====---+*+**=       
                                                   ****#####**++===+++*****++==--------===**#*+       
                                                 ..*######*++++=+++****+=========------===**##+..     
                                               ::######******++*****+=------======-----===*###***     
                                            .:-######*******###*+==---::::::-----------===+###***     
                                          ::+######*######***=--::::::.:..:::::--------===+####**     
                                        :-+############**+-:...................::---======+####**     
                                      --*#######*******=-.     .................:::-======+*###**     
                                    ==*#####*******#**+:...   .....................:::-===+*#****     
                                   .**###**++*+**##**=:.  ..........        .......::::-=++******     
                                 .=+###***+++***#**=-...   ......             .....:.::-=+****#**     
                                 :*###*+++++*##**+-::........              .........::-==**##****     
                               :+*##*+++****##*+-:-=-:.......    ..................::-===**#####*     
                               -*##*++**####*+-::::-=-:............................:-===+**#####*++   
                             =**##***###***+-.....::---:.........................:--====**######***   
                             +*##*#####**=:.........:::::.......................::-----+**#######**   
                           **#########*=:..........:::::::....................:::::::-+**########****.
                         ..########*+-:...........::::::::::.....   .......:::::::::=****#########*++.
                       ..**######*=-:............::::::::::::...    .....::::::::::-+****##**#####*.. 
                      .***####*+-:.:.............::::...:---::..   ...:::::::------=***####***####***.
                      .**###*+=:....      .......:::.....-=+=-:...:--:......::-----=***####****####**.
                    .:-*###*+=-:...      .................-+**==-=-:.........:---:-+***###******####*.
                    -**#%%#*+=:....     ...................:+##*+-:..........::--=+***####******####*.
                    -*#%%%#*=-....    ......................-**+-::..........:::-++**####****++*####*=
                  :-+#%%%%#+:.   ...........................:++=-:::.........::--=+**####*+**++**####*
                  =*#%%%%#+:.    ............................----:::::.......::--=+**####++++++**####*
                  =#%%%%#*-     ...............:::.........::::::::::::.......:-==+*#####++++++***###*
                -+*#%%%%*=..    ....................:::::::---::::::::......:::-=-=+**##*+++++++******
                =*#%%%%#=:.     ...:::::::..........:::--------::::::.......:::--=+==+++=======+**+++.
                +#%%%%%+-.     ....:::::::...........::-------::::::....:.::.::-==================+**.
                +#%%%%*:.      ....:::::::..........::::---::::::::... ....::-==================+=+   
              **#%%%%+:.        ....::::::::::::::::::::::::::....... ...::=+==================++     
              **#%%%+:.          ................:::::::::::::..........--=====++===========+++==     
              *#%%#+:..               ..............::::::::::........:::-====+++=========+++=-       
              ##%#+:...                 .....    .....................::-==++++==========++*:         
            ::###+-.:...                 .        ............       ..-=+***+=========++*==:         
           .**###*=.....                       ............          .-=+***++++++====++=-.           
           .**####*-..                ....  ............            .-=***+++++++==++*--              
           .**#####+...                 ...............           .:-+****+++++++++***                
            ::*#####=:.               ................        .   .-+******+++++***+::                
              **####*-:. .....       ...:............        .....:+*************+:.                  
              **#####+:.......      .....:...........       ...:=**##************=                    
              ***####*-:......      .....:...::......         :**###***********-.                     
              ***#####=::.....      ....::::::::.....     ..:.=*###***###******-                      
                +*####*::.         .....::::::::............::**##***###*****:                        
                =**###*=:.         ..........::..........::-=++*****###****                           
                  =*###+:.           .............:::-==++++++==++*******                             
                  =**##*=:.                 ....::-============+++*****++                             
                  -=+*##*-..               ...:----::.:::---=++++++***=                               
                    -**#*+-..            .::---:::....:::-=+++++++**+=-                               
                    .-=*#*+:.          ..----:....:::::-=+++++++**==:                                 
                      .**#*=..         :-=-::::::..:::-=+++++*----.                                   
                       --***:.        :--:...........:++++++--                                        
                         ***=.      .:--:  ........:=+++++*=                                          
                         ***+=.   ..--:........::---=+++*=:.                                          
                         ::**+-...:-:......:------==++*=:.                                            
                           **+=..:-:..::---=+++****....                                               
                           ..=+------===+*:........      

		Pygwarts core class, that serves as a super class for every pygwarts object. Every object, that
		inherited from Transmutable class might be either a single object, or be nested to other Transmutable
		object, forming a "mutable chain". This nesting means declaration of one Transmutable class in the
		namespace of the other Transmutable class. This nesting ensures that every nested class will become
		a nested layer object for current layer object, inferring the very top layer object, that will have
		access to all nested layers. Layers nesting means setting an object as an attribute of it's upper
		layer object, using it's declaration name. The mutable chain provides not only top-down layers
		access, but also a one way down-top access. That means for every nested layer there is a field
		"_UPPER_LAYER", which is used to access it's upper layer object namespace, so any upper layer
		member is reachable for nested layer. But it is only a direct access reachability, which means
		every nested layer may use dot notation to access members of upper layers, but not to access
		members of upper layers nested layers, that is a one way down-top access implies. This is the
		main feature of mutable chain, it is ideology, the concept. The other fundamental feature, that
		based on this concept, is a logging. It is hardcoded in pygwarts to use special name "loggy"
		as a LibraryContrib object to handle logging. It is not mandatory to declare LibraryContrib object
		as a "loggy" member, cause Transmutable will be able to find LibraryContrib object at any layer
		and reassign it as a "loggy", as it is possible to omit it too, so it will be bypassed. But it
		is very important to not use "loggy" keyword for any other objects, rather than LibraryContrib.
		The mutable chain might have different loggers for any nested layers, and all of them will be
		assigned and accessed as "loggy". One crucial moment is that any nested layer, that will encounter
		a LibraryContrib type, will not only initiate it and set it as it's "loggy" attribute, but also will
		escalate it the very top layer, if it doesn't have a LibraryContrib "loggy" attribute yet. If
		the very top layer already has it's "loggy" declared and initiated, any nested layer "loggy"
		declaration will be ignored, cause every "loggy" access has it's special preprocessing that
		first of all points to the very upper layer "loggy". If the very top layer already has it's "loggy"
		declared and initiated, but a nested layer has a LibraryContrib declared as some different name,
		this LibraryContrib will be initiated and assigned as "loggy" for current layer. If the very top
		layer doesn't has it's "loggy" declared and initiated, the very first nested layer LibraryContrib
		initiation encountered during mutable chain initiation will assigned this LibraryContrib as "loggy"
		for the very upper layer along with current layer. Any further LibraryContrib initiations for other
		nested layers will reassign this layer "loggy" if only were declared with different name.
		The mutable chain initiation is the process that include chain of initiations for every Transmutable
		class that nested to the other Transmutable class. It is consist of two phases:
			- first phase, the initial state of current Transmutable, that implies the "__init__" invoke
		itself. This phase considers the only positional argument it can receive - "upper_layer_link".
		That argument determines whether current Transmutable the very upper layer object or nested one.
			- second phase, the namespace scanning. Right after the first phase for current Transmutable,
		all members that are obtained from "__dir__" and a considered as Transmutable will be hooked and
		their phase one initiations will be invoked. It is also possible to encounter a LibraryContrib
		type member during this phase, so the events described above will occur. Such namespace scanning
		assumes a depth-first initiation for a mutable chain, cause every upper layer will invoke it's
		nested layer and proceed only after it's initiation complete. This phase also handles any Exceptions
		that might be raised during it's nested layer initiation, and also situations when somehow
		nested layer initiation results a not Transmutable object.
		After these two phases the very upper layer, as the only layer that might have keyword arguments
		provided for initiation, will process it and assign as fields. This keyword arguments might be
		any type and serves any purposes, as they doesn't impact a mutable chaining.
		As mutable chain initiation relies on hooking nested Transmutables in phase two, every Transmutable
		must have mandatory field "_MUTABLE_CHAIN_HOOK" - the main field that serves as Transmutable type
		recognition. Every nested Transmutable can only be found for upper layer by this field set to True.
		There is also one more mandatory and switchable hook "_UPPER_LAYER_HOOK", that serves as a flag
		whether current object is the very upper layer or not. By default it set to True, so for every
		Transmutable that find itself to be a nested layer, it must switch this hook to False, cause some
		things, like attribute searching, depends on this hook value.
		This super class for every pygwarts object original design implies simplicity and includes the
		only basic things that forms it's concept. Any inherited object's behavior supposed to be based
		on this concept. The main idea of a mutable chain strongly relies on declaration and it's order.
	"""


	_UPPER_LAYER				:Optional[Transmutable]
	_MUTABLE_CHAIN_HOOK	= True
	_UPPER_LAYER_HOOK		= True


	def __str__(self):

		"""
			Implements mutable chain naming representation. All upper layers for current Transmutable
			object will form a concatenation of names in dot notation, just like in accessing.
		"""

		if	not self._UPPER_LAYER_HOOK:

			return	f"{self._UPPER_LAYER}.{self.__class__.__name__}"
		return		self.__class__.__name__




	def __getattribute__(self, attr :str) -> object :

		"""
			This method implements the preprocessing of some crucial mutable chain members. The main
			job is a "loggy" attribute access handling. This is the only point where LibraryContrib
			object access might be mapped to the object that actually access it. As LibraryContrib
			object has special method "handover", every "loggy" access will be preceded with it's
			handover. Also, depending on current state of "loggy" member, if it is just encountered
			as a type, it will be initiated and also escalated to the very top layer. Any other than
			LibraryContrib entity, which is also can be determined by "CONTRIB_HOOK" field set to True,
			that discovered as a "loggy" member will cause TypeError.
		"""

		if	attr == "__qualname__" and attr not in dir(self) : return self.__class__.__name__


		_attr = super().__getattribute__(attr)


		if	attr == "loggy":
			match _attr:


				case	LibraryContrib(): return _attr.handover(self)

				case	type() if getattr(_attr, "CONTRIB_HOOK", None):		_attr = _attr()
				case	object() if getattr(_attr, "CONTRIB_HOOK", None):	_attr = _attr()()

				case _:	raise TypeError(f"\"loggy\" reserved for LibraryContrib, got {type(_attr)}")


			_attr.handover(self)
			_attr.CONTRIB_HOOK = False


			if	not self._UPPER_LAYER_HOOK : wingardium_leviosa(self, "loggy", _attr)
		return	_attr




	def __getattr__(self, attr :str) -> object :

		"""
			This method means __getattribute__ failed to obtain some attribute, so, as mutable chain
			structure suppose, every attribute will be searched to the very top layer. That means
			via "_UPPER_LAYER" accessing to the current Transmutable upper layer Transmutable, the
			attribute will be considered as a member of upper layer Transmutable. The truthy hook
			means the very top layer is reached and final AttributeError must be raised, cause
			desired attribute is not a part of mutable chain. There is the only exception for "loggy"
			attribute. If the very top layer is reached and there is no "loggy", special object
			"evanesce" will be returned to bypass "loggy" member necessity.
		"""

		if	attr == "loggy":
			if	self._UPPER_LAYER_HOOK : return evanesce


			# This is the point where some nested layer requested "loggy" from it's upper layer. That means
			# this nested layer had not yet encountered "loggy" object and handovered it. When "loggy" will
			# be finally found at some upper layer, built-in getattr function will put it through some
			# validation procedure, so it is guaranteed that result is a LibraryContrib object. For every
			# layer on the way to that LibraryContrib object, when descending back to a requester, handover
			# will ensure corresponding "loggy" assignment.
			_attr = getattr(self._UPPER_LAYER, attr)
			_attr.handover(self)


			return	_attr


		elif	not self._UPPER_LAYER_HOOK: return getattr(self._UPPER_LAYER, attr)
		raise	AttributeError(f"Attribute '{attr}' has no escalation to {self}")




	def __init__(self, upper_layer_link=None, **kwargs):


		# As mutable chain initiation is quiet simple and straightforward, hence there is no need in
		# logging during this process, and in terms of not messing up with handovers and so on, all
		# initiation phases goes silent, means with no logging, as it is probably useless.
		# Here starts first phase of initiation.
		match upper_layer_link:

			# This is the main case for mutable chain initiation. The "upper_layer_link" as Transmutable
			# itself means current Transmutable object is a nested layer for "upper_layer_link" and this
			# point was reached by "upper_layer_link" phase two of initiation, where current Transmutable
			# class was hooked and it's initiation was invoked.
			case Transmutable():

				self._UPPER_LAYER_HOOK = False
				self._UPPER_LAYER = upper_layer_link
				upper_layer_link.loggy.handover(self)


			# Following two cases covers mutable chain most upper layer initiation, as if LibraryContrib
			# object or class was passed as argument for initiation. As regular mutable chain initiation
			# suggests obtaining "loggy" via phase two of initiation, when it is encountered as declared
			# member, current two cases ensures an option of using some outer LibraryContrib entity.
			case LibraryContrib(): upper_layer_link.handover(self)
			case type() if getattr(upper_layer_link, "CONTRIB_HOOK", False):

				upper_layer_link().handover(self)
				self.loggy.CONTRIB_HOOK = False


			# Last two cases are regular case for most upper layer initiation, when "loggy" to be find
			# as any descent declared member in phase two of initiation, or even just omitted, and the
			# case when "upper_layer_link" neither Transmutable nor LibraryContrib entity, which is
			# mutable chain initiation violation, so a TypeError will be raised.
			case None:	pass
			case _:		raise TypeError(f"\"{upper_layer_link}\" is inappropriate upper layer link")




		# Here starts the second phase of initiation. It encompasses scanning the whole namespace
		# of the current Transmutable object, by the use of "__dir__" default method, instead of
		# built-in function "dir", which does not grant the iteration in order, that members were
		# declared. It is crucial in mutable chain initiation, cause, despite the original design
		# insist on mutable chain members independence, it is not strictly prohibited, so some
		# Transmutable's design must have the ability to rely on initiation in certain order to work.
		# The following scanning will search only Transmutable classes, which names not starts
		# with underscore and which has special fields named hooks that set to True. This is the
		# signal for current object that a nested layer Transmutable is encountered and must be
		# initiated as a member of a mutable chain.
		for inner in self.__dir__():
			if	not inner.startswith("_"):


				# It is crucial moment, that at this point the LibraryContrib class that declared as
				# "loggy", wherever it will be founded, will be handled very special way. As it is
				# mandatory goes via "getattr" built-in function, which invokes "__getattribute__"
				# method, the "loggy" name will trigger special handle for LibraryContrib class.
				# It is truly for the case when LibraryContrib is only declared as "loggy", so it
				# was never initiated yet. The declaration place is no matter, whether it is very
				# top layer or any nested one. The other case of "loggy" is a LibraryContrib object,
				# already initiated, is the most regular situation for any nested Transmutable that
				# was initiated by it's upper layer that already has "loggy" object assigned. Despite
				# this "loggy" was already assigned in first initiation phase for current Transmutable,
				# the fact it is will be accessed here, in second phase, and putted to the same processing
				# again, is not a redundant, but a additional layer of validation!
				layer_hook = getattr(self, inner)


				# This is the case when LibraryContrib is declared as any other name than "loggy".
				# It is possible for different mutable chain layers to have it's own "loggy" members,
				# so it is not only obvious for such members to be declared with different names, but
				# also very important, cause handling of "loggy" member has critical particularity:
				# if there will be more than one LibraryContrib class (means class with CONTRIB_HOOK
				# which is set to True by default) that will have different name, this object will
				# be set as loggy for current layer; if the same name "loggy" will be declared for
				# different layers, it is always the upper layer "loggy" will be returned as already
				# initiated, so other "loggy" declarations will be skipped.
				if	getattr(layer_hook, "CONTRIB_HOOK", None):
					setattr(self, "loggy", layer_hook)
					getattr(self, "loggy")
					continue


				# For a Transmutable the "_MUTABLE_CHAIN_HOOK" set to True is a trigger to consider
				# this "layer" as a Transmutable to be initiated. At this point any encountered
				# entity that has "_MUTABLE_CHAIN_HOOK" is considered to be a Transmutable type.
				elif	getattr(layer_hook, "_MUTABLE_CHAIN_HOOK", None):	layer = layer_hook
				else:	continue


				# The ending step for every mutable chain member candidate. At this point a "layer"
				# must be a Transmutable class which, when will be invoked with an "upper_layer_link"
				# as a current layer Transmutable object, will start it's own descent mutable chain
				# initiation, and will return a Transmutable object to be assigned as a nested layer
				# for current Transmutable object. Such assignment is just a setting an attribute
				# that points to an object, using it's declaration name "inner". Any possible exceptions,
				# that might occur during nested Transmutable initiation, will be caught and basically
				# skipped, means the global initiation of mutable chain will not be interrupted, but
				# current layer will try to log corresponding error message. If somehow "layer"
				# initiation will result not a Transmutable object, and in case "layer" initiation
				# caused Exception raise, such "inner" member will be deleted from current Transmutable
				# namespace, to prevent undefined behavior for improperly initiated mutable chain members,
				# cause it remained as a type.
				try:	mutable = layer(self)
				except	Exception as E : self.loggy.error(f"{inner} nesting caused {patronus(E)}")
				else:
					if	isinstance(mutable, Transmutable):
						setattr(self, inner, mutable)
						continue


					self.loggy.error(f"{inner} nesting caused not Transmutable mutation")
				delattr(self.__class__, inner)




		# The last step for phase two of initiation is an optional keyword processing. That stands
		# for maintaining any keyword arguments at the very top layer of mutable chain, cause
		# probably there is no way any nested layer could be initiated with any arguments other
		# than "upper_layer_link". It is also the place where such keyword arguments are accessible
		# for any descent mutable chain member. The purpose of such keyword arguments is the same
		# as for declarable fields, so fields could be overwritten, but only on the very top layer,
		# any descent fields are out of touch and are reachable at the first place for members
		# they are declared under. But the very top layer arguments are still reachable for any
		# member, even with it's own fields with same names, via the "_UPPER_LAYER" chaining calls,
		# so the very top layer can provide "defaults". Anyway, no descent member should rely on
		# such keyword arguments in their initiation phases, cause this keyword arguments will become
		# a part of mutable chain in the very last moment.
		for k,v in kwargs.items() : setattr(self, k,v)







